"""Class for getting phonetic pronunciations for tokenized text"""
import itertools
import logging
import sqlite3
import typing
from pathlib import Path

from gruut.const import PHONEMES_TYPE

# -----------------------------------------------------------------------------

_LOGGER = logging.getLogger("gruut.phonemize")

ROLE_TO_PHONEMES = typing.Dict[str, PHONEMES_TYPE]

WORD_TRANSFORM_TYPE = typing.Callable[[str], str]


# -----------------------------------------------------------------------------


class SqlitePhonemizer:
    """Phonemizes text using a lexicon from a sqlite database"""

    DEFAULT_ROLE: str = ""

    def __init__(
        self,
        db_conn: sqlite3.Connection,
        lexicon: typing.Optional[typing.Dict[str, ROLE_TO_PHONEMES]] = None,
        g2p_model: typing.Optional[typing.Dict[str, typing.Union[str, Path]]] = None,
        word_transform_funcs: typing.Optional[
            typing.Iterable[WORD_TRANSFORM_TYPE]
        ] = None,
        casing_func: typing.Optional[WORD_TRANSFORM_TYPE] = None,
    ):
        self.db_conn = db_conn

        # word -> role -> [phonemes]
        self.lexicon = lexicon if lexicon is not None else {}

        # [functions]
        self.word_transform_funcs = word_transform_funcs or []

        self.casing_func = casing_func

    def __call__(
        self, word: str, role: typing.Optional[str] = None, do_transforms: bool = True
    ) -> typing.Optional[PHONEMES_TYPE]:
        # Look up in cache first
        if self.casing_func is not None:
            word = self.casing_func(word)

        role_to_word = self.lexicon.get(word)

        if role_to_word is not None:
            if role is not None:
                # Exact role
                phonemes = role_to_word.get(role)
                if phonemes is not None:
                    return phonemes

            # Default role
            phonemes = role_to_word.get(SqlitePhonemizer.DEFAULT_ROLE)
            if phonemes is not None:
                return phonemes

            # Any role
            if role_to_word:
                return next(iter(role_to_word.values()))

            # Not in lexicon (or database) for sure because role_to_word was present.
            return None

        transforms = self.word_transform_funcs
        if not do_transforms:
            # No transforms
            transforms = []

        for transform_func in itertools.chain([None], transforms):
            if transform_func is not None:
                lookup_word = transform_func(word)
            else:
                # No transform
                lookup_word = word

            if not lookup_word:
                continue

            # Load pronunciations for word from database.
            cursor = self.db_conn.execute(
                "SELECT role, phonemes FROM word_phonemes WHERE word = ? ORDER BY pron_order",
                (lookup_word,),
            )

            for row in cursor:
                if role_to_word is None:
                    # Create new lexicon entry for original word
                    role_to_word = {}
                    self.lexicon[word] = role_to_word

                db_role, db_phonemes = row[0], row[1].split()

                if db_role not in role_to_word:
                    role_to_word[db_role] = db_phonemes

            if role_to_word is not None:
                # Link to transformed word
                self.lexicon[lookup_word] = self.lexicon[word]

                # Successfully looked up in the database
                return self(word, role=role)

        # Not in lexicon
        return None



"""gruut module"""
import itertools
import logging
import re
import sqlite3
import threading
import typing
from enum import Enum
from pathlib import Path

from gruut.const import KNOWN_LANGS, TextProcessorSettings
from gruut.resources import _DIR, _PACKAGE
from gruut.text_processor import Sentence, TextProcessor
from gruut.utils import resolve_lang

# -----------------------------------------------------------------------------

_LOGGER = logging.getLogger(_PACKAGE)

__version__ = (_DIR / "VERSION").read_text(encoding="utf-8").strip()
__author__ = "Michael Hansen (synesthesiam)"
__all__ = [
    "sentences",
    "is_language_supported",
    "get_supported_languages",
    "TextProcessor",
    "TextProcessorSettings",
]

# -----------------------------------------------------------------------------

_LOCAL = threading.local()
_PROCESSORS_LOCK = threading.RLock()


def sentences(
    text: str,
    lang: str = "en_US",
    ssml: bool = False,
    espeak: bool = False,
    major_breaks: bool = True,
    minor_breaks: bool = True,
    punctuations: bool = True,
    explicit_lang: bool = True,
    phonemes: bool = True,
    break_phonemes: bool = True,
    pos: bool = True,
    **process_args,
) -> typing.Iterable[Sentence]:
    """
    Process text and return sentences

    Args:
        text: input text or SSML (ssml=True)
        lang: default language of input text
        ssml: True if input text is SSML
        espeak: True if eSpeak phonemes should be used
        major_breaks: False if no sentence-breaking symbols in output
        minor_breaks: False if no phrase-breaking symbols in output
        punctuations: False if no word-surrounding symbols in output
        **process_args: keyword arguments passed to TextProcessor.process

    Returns:
        sentences: iterable of Sentence objects

    """
    model_prefix = "" if (not espeak) else "espeak"

    with _PROCESSORS_LOCK:
        if not hasattr(_LOCAL, "processors"):
            _LOCAL.processors = {}

        text_processor = _LOCAL.processors.get(model_prefix)
        if text_processor is None:
            text_processor = TextProcessor(default_lang=lang, model_prefix=model_prefix)
            _LOCAL.processors[model_prefix] = text_processor

    assert text_processor is not None
    graph, root = text_processor(text, lang=lang, ssml=ssml, **process_args)

    yield from text_processor.sentences(
        graph,
        root,
        major_breaks=major_breaks,
        minor_breaks=minor_breaks,
        punctuations=punctuations,
        explicit_lang=explicit_lang,
        phonemes=phonemes,
        break_phonemes=break_phonemes,
        pos=pos,
    )


# -----------------------------------------------------------------------------


def is_language_supported(lang: str) -> bool:
    """True if gruut supports lang"""
    return resolve_lang(lang) in KNOWN_LANGS


def get_supported_languages() -> typing.Set[str]:
    """Set of supported gruut languages"""
    return set(KNOWN_LANGS)



#!/usr/bin/env python3
"""Tests for French"""
import typing
import unittest

from gruut import sentences

# # https://www.commeunefrancaise.com/blog/la-liaison

# # After a determiner
# Un enfant.
# Les arbres.
# Deux amis.
# Ton excellent vin.
# Ces autres voyages.

# # Pronoun + verb
# On est là!
# Elles ont faim!
# Vous êtes sûrs?
# Tu nous entends.
# Je les adore.

# # Adjective + noun
# J’ai des petites oreilles.
# Michel est un grand ami.
# Je regarde la télé sur un petit écran.
# C’est un ancien élève.

# # After short prepositions, and “très”
# C’est très amusant!
# Je vis en Amérique.
# Ils sont chez eux.
# J’arrive dans une minute.

# # Others
# Un bâtiment est en vue de l’île.
# Sa vie n’était pas en danger.
# Amalia est en danger.
# C`est incroyable!


class FrenchTestCase(unittest.TestCase):
    """Test cases for French"""

    def test_liason_after_determiner(self):
        """Test liason after determiner (e.g., le)"""
        self._without_and_with_liason("Les arbres", "Les", ["l", "e"], ["l", "e", "z"])

    def test_liason_adjective_noun(self):
        """Test liason between adjective and noun"""
        self._without_and_with_liason(
            "J’ai des petites oreilles.",
            "petites",
            ["p", "ə", "t", "i", "t"],
            ["p", "ə", "t", "i", "t", "z"],
        )

    def test_liason_pronoun_verb(self):
        """Test liason between pronoun and verb"""
        self._without_and_with_liason("On est là!", "On", ["ɔ̃"], ["ɔ̃", "n"])

    def test_liason_tres(self):
        """Test liason with très"""
        self._without_and_with_liason(
            "C’est très amusant!", "très", ["t", "ʁ", "ɛ"], ["t", "ʁ", "ɛ", "z"]
        )

    def _without_and_with_liason(
        self,
        text: str,
        liason_word: str,
        without_phonemes: typing.List[str],
        with_phonemes: typing.List[str],
    ):
        """Get pronunciation of a sentence with and without liason enabled"""

        # Verify no liason
        sentence = next(iter(sentences(text, lang="fr_FR", post_process=False)))
        word = next(w for w in sentence if w.text == liason_word)
        self.assertEqual(word.phonemes, without_phonemes)

        # Verify liason
        sentence = next(iter(sentences(text, lang="fr_FR", post_process=True)))
        word = next(w for w in sentence if w.text == liason_word)
        self.assertEqual(word.phonemes, with_phonemes)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Setup file for gruut"""
from collections import defaultdict
from pathlib import Path

import setuptools

this_dir = Path(__file__).parent
module_dir = this_dir / "gruut"

# -----------------------------------------------------------------------------

# Load README in as long description
long_description: str = ""
readme_path = this_dir / "README.md"
if readme_path.is_file():
    long_description = readme_path.read_text(encoding="utf-8")

requirements = []
requirements_path = this_dir / "requirements.txt"
if requirements_path.is_file():
    with open(requirements_path, "r", encoding="utf-8") as requirements_file:
        requirements = requirements_file.read().splitlines()

version_path = module_dir / "VERSION"
with open(version_path, "r", encoding="utf-8") as version_file:
    version = version_file.read().strip()

# x.y.z -> x.0.0
base_version = ".".join(version.split(".")[:-2] + ["0", "0"])

# -----------------------------------------------------------------------------
# extras_require
# -----------------------------------------------------------------------------

# dependency => [tags]
extras = {
    "hazm~=0.7.0": ["fa"],
    "conllu>=4.4": ["train"],
    "rapidfuzz>=1.4.1": ["train"],
    "aeneas~=1.7.3.0": ["align"],  # requires numpy to install
    "pydub~=0.24.1": ["align"],
    "mishkal~=0.4.0": ["ar"],
    "codernitydb3~=0.6.0": ["ar"],
    "phonetisaurus~=0.3.0": ["g2p"],
}

# Create language-specific extras
for lang in [
    "ar",
    "cs",
    "de",
    "es",
    "fa",
    "fr",
    "it",
    "lb",
    "nl",
    "pt",
    "ru",
    "sv",
    "sw",
]:
    extras[f"gruut_lang_{lang}~={base_version}"] = [lang]

# Add "all" tag
for tags in extras.values():
    tags.append("all")

# Invert for setup
extras_require = defaultdict(list)
for dep, tags in extras.items():
    for tag in tags:
        extras_require[tag].append(dep)


# -----------------------------------------------------------------------------

include_files = {"lexicon.db", "model.crf"}

data_dir = module_dir / "data"
data_files = [
    str(f.relative_to(module_dir))
    for f in data_dir.rglob("*")
    if f.is_file() and (f.name in include_files)
]

setuptools.setup(
    name="gruut",
    description="A tokenizer, text cleaner, and phonemizer for many human languages.",
    version=version,
    author="Michael Hansen",
    author_email="mike@rhasspy.org",
    url="https://github.com/rhasspy/gruut",
    packages=setuptools.find_packages(),
    package_data={"gruut": data_files + ["VERSION", "py.typed"]},
    install_requires=requirements,
    extras_require={
        ':python_version<"3.7"': ["dataclasses", "types-dataclasses"],
        ':python_version<"3.9"': ["importlib_resources"],
        **extras_require,
    },
    entry_points={"console_scripts": ["gruut = gruut.__main__:main"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
)



"""Shared classes, types, and enums"""
import itertools
import operator
import re
import typing
import xml.etree.ElementTree as etree
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum

import babel
import babel.core
import babel.numbers

# alias -> full language name
LANG_ALIASES = {
    "ar": "ar",
    "cs": "cs-cz",
    "de": "de-de",
    "en": "en-us",
    "en-gb": "en-us",
    "es": "es-es",
    "es-mx": "es-es",
    "fa": "fa",
    "fr": "fr-fr",
    "it": "it-it",
    "lb-lb": "lb",
    "nl": "nl",
    "nl-nl": "nl",
    "pt-br": "pt",
    "ru": "ru-ru",
    "sv": "sv-se",
    "sw": "sw",
    "zh": "zh-cn",
}

ENGLISH_LANGS = {"en-us", "en-gb"}

# Languages that are expected to have a model directory
KNOWN_LANGS = set(itertools.chain(ENGLISH_LANGS, LANG_ALIASES.values()))


try:
    # Python >= 3.7
    REGEX_PATTERN = re.Pattern  # type: ignore
    REGEX_MATCH = re.Match  # type: ignore
    REGEX_TYPE = typing.Union[str, re.Pattern]  # type: ignore
except AttributeError:
    # Python 3.6
    REGEX_PATTERN = typing.Pattern  # type: ignore
    REGEX_MATCH = typing.Match  # type: ignore
    REGEX_TYPE = typing.Union[str, typing.Pattern]  # type: ignore


# Phonemes for a single word
PHONEMES_TYPE = typing.Sequence[str]

# Type of nodes in a text graph
NODE_TYPE = int

# Property used to hold node data in text graph
DATA_PROP = "data"


class GraphType:
    """Type wrapper for networkx graph"""

    nodes: typing.Dict[NODE_TYPE, typing.Dict[typing.Any, typing.Any]]
    """Get node data for the graph"""

    def add_node(self, node: NODE_TYPE, **kwargs):
        """Add a new node to the graph"""
        pass

    def add_edge(self, src: NODE_TYPE, dst: NODE_TYPE):
        """Add a new edge to the graph"""
        pass

    def out_degree(self, node: NODE_TYPE) -> int:
        """Get number of outgoing edges from a node"""
        pass

    def successors(self, node: NODE_TYPE) -> typing.Iterable[NODE_TYPE]:
        """Yield nodes on outgoing edges"""
        pass

    def predecessors(self, node: NODE_TYPE) -> typing.Iterable[NODE_TYPE]:
        """Yield nodes from incoming edges"""
        pass

    def out_edges(
        self, node: NODE_TYPE
    ) -> typing.Iterable[typing.Tuple[NODE_TYPE, NODE_TYPE]]:
        """Yield outgoing edges from a node"""
        pass

    def add_edges_from(
        self, edges: typing.Iterable[typing.Tuple[NODE_TYPE, NODE_TYPE]]
    ):
        """Add edges from iterable"""
        pass

    def remove_edges_from(
        self, edges: typing.Iterable[typing.Tuple[NODE_TYPE, NODE_TYPE]]
    ):
        """Remove edges from iterable"""
        pass

    def __len__(self) -> int:
        """Get number of nodes in the graph"""
        pass


# -----------------------------------------------------------------------------

DEFAULT_SPLIT_PATTERN = re.compile(r"(\s+)")

NORMALIZE_WHITESPACE_PATTERN = re.compile(r"\s+")
SURROUNDING_WHITESPACE_PATTERN = re.compile(r"^(\s*)\S+(\s*)$")
HAS_DIGIT_PATTERN = re.compile(r"[0-9]")


@dataclass
class Time:
    """Parsed time from text"""

    hours: int
    minutes: int = 0

    period: typing.Optional[str] = None
    """A.M. or P.M."""


# -----------------------------------------------------------------------------


class InterpretAs(str, Enum):
    """Supported options for interpret-as attribute of <say-as>"""

    SPELL_OUT = "spell-out"
    """Word should be spelled out (abc = a b c)"""

    DATE = "date"
    """Word should be interpreted as a date"""

    NUMBER = "number"
    """Word should be interpreted as a number"""

    CURRENCY = "currency"
    """Word should be interpreted as an amount of currency"""

    TIME = "time"
    """Word should be interpreted as a time on the clock"""

    WORD = "word"
    """Interpret as regular word"""


class InterpretAsFormat(str, Enum):
    """Supported options for format attribute of <say-as>"""

    NUMBER_CARDINAL = "cardinal"
    """Cardinal version of number (1 = one)"""

    NUMBER_ORDINAL = "ordinal"
    """Ordinal version of number (1 = first)"""

    NUMBER_DIGITS = "digits"
    """Number as digits (12 = one two)"""

    NUMBER_YEAR = "year"
    """Number as a year (2021 = twenty twenty-one)"""

    # Date formats
    # d = day
    # m = month
    # y = year
    # o = ordinal day ("first" instead of "one")
    DATE_DMY = "dmy"
    DATE_MDY = "mdy"
    DATE_YMD = "ymd"
    DATE_DMY_ORDINAL = "omy"
    DATE_MDY_ORDINAL = "moy"
    DATE_YMD_ORDINAL = "ymo"
    DATE_YM = "ym"
    DATE_MY = "my"
    DATE_MD = "md"
    DATE_MD_ORDINAL = "mo"
    DATE_DM_ORDINAL = "om"
    DATE_Y = "y"


class BreakType(str, Enum):
    """Types of sentence breaks"""

    MINOR = "minor"
    """Break between phrases"""

    MAJOR = "major"
    """Break between sentences"""


class WordRole(str, Enum):
    """Role of a word. Used to disambiguate pronunciations."""

    DEFAULT = ""
    """Use default word pronunciation"""

    LETTER = "gruut:letter"
    """Word should be pronounced as a letter (a = /eɪ/ instead of /ə/)"""


class SSMLParsingState(int, Enum):
    """Current state of SSML parsing"""

    DEFAULT = 0

    IN_WORD = 1
    """Inside <w> or <token>"""

    IN_LEXICON = 2
    """Inside <lexicon>"""

    IN_LEXICON_GRAPHEME = 3
    """Inside <lexicon><grapheme>..."""

    IN_LEXICON_PHONEME = 4
    """Inside <lexicon><phoneme>..."""


@dataclass
class InlineLexicon:
    """SSML lexicon defined inline (not standards compliant)"""

    lexicon_id: str
    alphabet: str = ""

    # word -> role -> [phoneme]
    words: typing.Dict[str, typing.Dict[str, PHONEMES_TYPE]] = field(
        default_factory=dict
    )


@dataclass
class Lexeme:
    """Entry of an inline lexicon"""

    grapheme: str = ""
    phonemes: typing.Optional[PHONEMES_TYPE] = None
    roles: typing.Optional[typing.Set[str]] = None


@dataclass
class Node:
    """Base class of all text processing graph nodes"""

    node: NODE_TYPE
    element: typing.Optional[etree.Element] = None
    voice: str = ""
    lang: str = ""
    implicit: bool = False


@dataclass
class IgnoreNode(Node):
    """Node should be ignored"""

    pass


@dataclass
class BreakNode(Node):
    """Represents a user-specified break"""

    time: str = ""
    """Length of break in seconds (123s) or milliseconds (123ms)"""

    def get_milliseconds(self) -> int:
        """Get number of milliseconds from the time string"""
        if self.time.endswith("ms"):
            return int(self.time[:-2])

        if self.time.endswith("s"):
            return int(float(self.time[:-1]) * 1000)

        return 0


@dataclass
class MarkNode(Node):
    """Represents a user-specified mark"""

    name: str = ""
    """Name of the mark"""


@dataclass
class WordNode(Node):
    """Represents a single word"""

    text: str = ""
    text_with_ws: str = ""
    interpret_as: typing.Union[str, InterpretAs] = ""
    format: typing.Union[str, InterpretAsFormat] = ""

    number: typing.Optional[Decimal] = None
    date: typing.Optional[datetime] = None
    currency_symbol: typing.Optional[str] = None
    currency_name: typing.Optional[str] = None
    time: typing.Optional[Time] = None

    role: typing.Union[str, WordRole] = WordRole.DEFAULT
    pos: typing.Optional[str] = None
    phonemes: typing.Optional[typing.Sequence[str]] = None

    in_lexicon: typing.Optional[bool] = None
    lexicon_ids: typing.Optional[typing.Sequence[str]] = None

    # Assume yes until proven otherwise
    is_maybe_number: bool = True
    is_maybe_date: bool = True
    is_maybe_currency: bool = True
    is_maybe_time: bool = True

    is_from_broken_word: bool = False


@dataclass
class BreakWordNode(Node):
    """Represents a major/minor break in the text"""

    break_type: typing.Union[str, BreakType] = ""
    text: str = ""
    text_with_ws: str = ""


@dataclass
class PunctuationWordNode(Node):
    """Represents a punctuation marker in the text"""

    text: str = ""
    text_with_ws: str = ""


@dataclass
class SentenceNode(Node):
    """Represents a sentence with WordNodes under it"""

    pass


@dataclass
class ParagraphNode(Node):
    """Represents a paragraph with SentenceNodes under it"""

    pass


@dataclass
class SpeakNode(Node):
    """Top-level node for SSML"""

    pass


# -----------------------------------------------------------------------------


@dataclass
class Word:
    """Processed word from a Sentence"""

    idx: int
    """Zero-based index of word in sentence"""

    text: str
    """Text with normalized whitespace"""

    text_with_ws: str
    """Text with original whitespace"""

    leading_ws: str = ""
    """Whitespace before text"""

    trailing_ws: str = ""
    """Whitespace after text"""

    sent_idx: int = 0
    """Zero-based index of sentence in paragraph"""

    par_idx: int = 0
    """Zero-based index of paragraph in document"""

    lang: str = ""
    """Language code"""

    voice: str = ""
    """Voice (from SSML)"""

    pos: typing.Optional[str] = None
    """Part of speech (None if not set)"""

    phonemes: typing.Optional[typing.Sequence[str]] = None
    """List of phonemes (None if not set)"""

    is_major_break: bool = False
    """True if word is a major break (separates sentences)"""

    is_minor_break: bool = False
    """True if word is a minor break (separates phrases)"""

    is_punctuation: bool = False
    """True if word is punctuation that surrounds a spoken word (quotes, etc.)"""

    is_break: typing.Optional[bool] = None
    """True if major or minor break"""

    is_spoken: typing.Optional[bool] = None
    """True if word is something that would be spoken during reading (not punctuation or break)"""

    pause_before_ms: int = 0
    """Milliseconds to pause before this word"""

    pause_after_ms: int = 0
    """Milliseconds to pause after this word"""

    marks_before: typing.Optional[typing.List[str]] = None
    """User-defined marks that occur before this word"""

    marks_after: typing.Optional[typing.List[str]] = None
    """User-defined marks that occur after this word"""

    def __post_init__(self):
        if self.is_break is None:
            self.is_break = self.is_major_break or self.is_minor_break

        if self.is_spoken is None:
            self.is_spoken = not (self.is_punctuation or self.is_break)

        self.leading_ws, self.trailing_ws = default_get_whitespace(self.text_with_ws)


@dataclass
class Sentence:
    """Processed sentence from a document"""

    idx: int
    """Zero-based index of sentence in paragraph"""

    text: str
    """Text with normalized whitespace"""

    text_with_ws: str
    """Text with original whitespace"""

    text_spoken: str
    """Text with only spoken words and normalized whitespace"""

    par_idx: int = 0
    """Zero-based index of paragraph in document"""

    lang: str = ""
    """Language code"""

    voice: str = ""
    """Voice (from SSML)"""

    words: typing.List[Word] = field(default_factory=list)
    """Words in the sentence"""

    pause_before_ms: int = 0
    """Milliseconds to pause before this sentence"""

    pause_after_ms: int = 0
    """Milliseconds to pause after this sentence"""

    marks_before: typing.Optional[typing.List[str]] = None
    """User-defined marks that occur before this sentence"""

    marks_after: typing.Optional[typing.List[str]] = None
    """User-defined marks that occur after this sentence"""

    def __iter__(self):
        """Iterates over words"""
        return iter(self.words)

    def __len__(self):
        """Number of words"""
        return len(self.words)

    def __getitem__(self, key):
        """Gets word by index"""
        return self.words[key]


# -----------------------------------------------------------------------------


class LookupPhonemes:
    """Look up phonemes for word/role in a lexicon"""

    def __call__(
        self, word: str, role: typing.Optional[str] = None, do_transforms: bool = True
    ) -> typing.Optional[PHONEMES_TYPE]:
        pass


class GuessPhonemes:
    """Guess phonemes for word/role"""

    def __call__(
        self, word: str, role: typing.Optional[str] = None
    ) -> typing.Optional[PHONEMES_TYPE]:
        pass


class GetPartsOfSpeech:
    """Get part of speech tags for words"""

    def __call__(self, words: typing.Sequence[str]) -> typing.Sequence[str]:
        pass


class PostProcessSentence:
    """Post-process each sentence node after tokenization/phonemization"""

    def __call__(
        self, graph: GraphType, sentence_node: SentenceNode, settings: typing.Any,
    ):
        pass


@dataclass
class EndElement:
    """Wrapper for end of an XML element (used in TextProcessor)"""

    element: etree.Element


# -----------------------------------------------------------------------------


def has_digit(s: str) -> bool:
    """True if string contains at least one digit"""
    return HAS_DIGIT_PATTERN.search(s) is not None


DEFAULT_WORD_PATTERN = re.compile(r"(\s*\S+(?:\s+|$))")


def default_split_words(s: str) -> typing.Iterable[str]:
    """Split text on whitespace"""
    yield from filter(None, DEFAULT_WORD_PATTERN.findall(s))


def default_get_whitespace(s: str) -> typing.Tuple[str, str]:
    """Returns leading and trailing whitespace of a string"""
    leading_ws, trailing_ws = "", ""
    match = SURROUNDING_WHITESPACE_PATTERN.match(s)
    if match is not None:
        leading_ws, trailing_ws = match.groups()

    return leading_ws, trailing_ws


def default_normalize_whitespace(s: str) -> str:
    """Replace multiple spaces with single space"""
    return NORMALIZE_WHITESPACE_PATTERN.sub(" ", s.strip())


def maybe_compile_regex(
    str_or_pattern: typing.Union[str, REGEX_PATTERN]
) -> REGEX_PATTERN:
    """Compile regex pattern if it's a string"""
    if isinstance(str_or_pattern, REGEX_PATTERN):
        return str_or_pattern

    assert isinstance(str_or_pattern, str)

    return re.compile(str_or_pattern)


# -----------------------------------------------------------------------------


@dataclass
class TextProcessorSettings:
    """Language specific settings for text processing"""

    lang: str
    """Language code that these settings apply to (e.g., en_US)"""

    # Whitespace/tokenization
    split_words: typing.Callable[[str], typing.Iterable[str]] = default_split_words
    """Split text into words and separators"""

    join_str: str = " "
    """String used to combine text from words"""

    keep_whitespace: bool = True
    """True if original whitespace should be retained"""

    is_non_word: typing.Optional[typing.Callable[[str], bool]] = None
    """Returns true if text is not a word (and should be ignored in final output)"""

    get_whitespace: typing.Callable[
        [str], typing.Tuple[str, str]
    ] = default_get_whitespace
    """Returns leading, trailing whitespace from a string"""

    normalize_whitespace: typing.Callable[[str], str] = default_normalize_whitespace
    """Normalizes whitespace in a string"""

    # Punctuations
    begin_punctuations: typing.Optional[typing.Set[str]] = None
    """Strings that should be split off from the beginning of a word"""

    begin_punctuations_pattern: typing.Optional[REGEX_TYPE] = None
    """Regex that overrides begin_punctuations"""

    end_punctuations: typing.Optional[typing.Set[str]] = None
    """Strings that should be split off from the end of a word"""

    end_punctuations_pattern: typing.Optional[REGEX_TYPE] = None
    """Regex that overrides end_punctuations"""

    # Replacements/abbreviations
    replacements: typing.Sequence[typing.Tuple[REGEX_TYPE, str]] = field(
        default_factory=list
    )
    """Regex, replacement template pairs that are applied in order right after tokenization on each word"""

    abbreviations: typing.Dict[REGEX_TYPE, str] = field(default_factory=dict)
    """Regex, replacement template pairs that may expand words after minor breaks are matched"""

    spell_out_words: typing.Dict[str, str] = field(default_factory=dict)
    """Written form, spoken form pairs that are applied with interpret-as="spell-out" in <say-as>"""

    # Breaks
    major_breaks: typing.Set[str] = field(default_factory=set)
    """Set of strings that occur at the end of a word and should break apart sentences."""

    major_breaks_pattern: typing.Optional[REGEX_TYPE] = None
    """Regex that overrides major_breaks"""

    minor_breaks: typing.Set[str] = field(default_factory=set)
    """Set of strings that occur at the end of a word and should break apart phrases."""

    minor_breaks_pattern: typing.Optional[REGEX_TYPE] = None
    """Regex that overrides minor_breaks"""

    word_breaks: typing.Set[str] = field(default_factory=set)
    word_breaks_pattern: typing.Optional[REGEX_TYPE] = None
    """Regex that overrides word_breaks"""

    # Numbers
    is_maybe_number: typing.Optional[typing.Callable[[str], bool]] = has_digit
    """True if a word may be a number (parsing will be attempted)"""

    get_ordinal: typing.Optional[typing.Callable[[str], typing.Optional[int]]] = None
    """Returns integer value of an ordinal string (e.g., 1st -> 1) or None if not an ordinal"""

    babel_locale: typing.Optional[str] = None
    """Locale used to parse numbers/dates/currencies (defaults to lang)"""

    num2words_lang: typing.Optional[str] = None
    """Language used to verbalize numbers (defaults to lang)"""

    # Currency
    default_currency: str = "USD"
    """Currency name to use when interpret-as="currency" but no currency symbol is present"""

    currencies: typing.MutableMapping[str, str] = field(default_factory=dict)
    """Mapping from currency symbol ($) to currency name (USD)"""

    currency_symbols: typing.Sequence[str] = field(default_factory=list)
    """Ordered list of currency symbols (decreasing length)"""

    is_maybe_currency: typing.Optional[typing.Callable[[str], bool]] = has_digit
    """True if a word may be an amount of currency (parsing will be attempted)"""

    # Dates
    dateparser_lang: typing.Optional[str] = None
    """Language used to parse dates (defaults to lang)"""

    is_maybe_date: typing.Optional[typing.Callable[[str], bool]] = has_digit
    """True if a word may be a date (parsing will be attempted)"""

    default_date_format: typing.Union[
        str, InterpretAsFormat
    ] = InterpretAsFormat.DATE_MDY_ORDINAL
    """Format used to verbalize a date unless set with the format attribute of <say-as>"""

    # Times
    is_maybe_time: typing.Optional[typing.Callable[[str], bool]] = has_digit
    """True if a word may be a clock time (parsing will be attempted)"""

    parse_time: typing.Optional[typing.Callable[[str], typing.Optional[Time]]] = None
    """Parse word text into a Time object or None"""

    verbalize_time: typing.Optional[
        typing.Callable[[Time], typing.Iterable[str]]
    ] = None
    """Convert Time to words"""

    # Part of speech (pos) tagging
    get_parts_of_speech: typing.Optional[GetPartsOfSpeech] = None
    """Optional function to get part of speech for a word"""

    # Initialisms (e.g, TTS or T.T.S.)
    is_initialism: typing.Optional[typing.Callable[[str], bool]] = None
    """True if a word is an initialism (will be split with split_initialism)"""

    split_initialism: typing.Optional[
        typing.Callable[[str], typing.Sequence[str]]
    ] = None
    """Function to break apart an initialism into multiple words (called if is_initialism is True)"""

    # Phonemization
    lookup_phonemes: typing.Optional[LookupPhonemes] = None
    """Optional function to look up phonemes for a word/role (without guessing)"""

    guess_phonemes: typing.Optional[GuessPhonemes] = None
    """Optional function to guess phonemes for a word/role"""

    # Pre/post-processing
    pre_process_text: typing.Optional[typing.Callable[[str], str]] = None
    """Optional function to process text during tokenization"""

    post_process_sentence: typing.Optional[PostProcessSentence] = None
    """Optional function to post-process each sentence in the graph before post_process_graph"""

    def __post_init__(self):
        # Languages/locales
        if self.babel_locale is None:
            if "-" in self.lang:
                # en-us -> en_US
                lang_parts = self.lang.split("-", maxsplit=1)
                self.babel_locale = "_".join(
                    [lang_parts[0].lower(), lang_parts[1].upper()]
                )
            else:
                self.babel_locale = self.lang

        if self.num2words_lang is None:
            self.num2words_lang = self.babel_locale

        if self.dateparser_lang is None:
            # en_US -> en
            self.dateparser_lang = self.babel_locale.split("_")[0]

        # Pre-compiled regular expressions
        self.replacements = [
            (maybe_compile_regex(pattern), template)
            for pattern, template in self.replacements
        ]

        compiled_abbreviations = {}
        for pattern, template in self.abbreviations.items():
            if isinstance(pattern, str):
                if not pattern.endswith("$") and self.major_breaks:
                    # Automatically add optional major break at the end
                    break_pattern_str = "|".join(
                        re.escape(b) for b in self.major_breaks
                    )
                    pattern = (
                        f"{pattern}(?P<break>{break_pattern_str})?(?P<whitespace>\\s*)$"
                    )
                    template += r"\g<break>\g<whitespace>"

                pattern = re.compile(pattern)

            compiled_abbreviations[pattern] = template

        self.abbreviations = compiled_abbreviations

        # Strings that should be separated from words, but do not cause any breaks
        if (self.begin_punctuations_pattern is None) and self.begin_punctuations:
            pattern_str = "|".join(re.escape(b) for b in self.begin_punctuations)

            # Match begin_punctuations only at start a word
            self.begin_punctuations_pattern = f"^({pattern_str})"

        if self.begin_punctuations_pattern is not None:
            self.begin_punctuations_pattern = maybe_compile_regex(
                self.begin_punctuations_pattern
            )

        if (self.end_punctuations_pattern is None) and self.end_punctuations:
            pattern_str = "|".join(re.escape(b) for b in self.end_punctuations)

            # Match end_punctuations only at end of a word
            self.end_punctuations_pattern = f"({pattern_str})$"

        if self.end_punctuations_pattern is not None:
            self.end_punctuations_pattern = maybe_compile_regex(
                self.end_punctuations_pattern
            )

        # Major breaks (split sentences)
        if (self.major_breaks_pattern is None) and self.major_breaks:
            pattern_str = "|".join(re.escape(b) for b in self.major_breaks)

            # Match major break with either whitespace at the end or at the end of the text
            # Allow for multiple punctuation symbols (e.g., !?)
            self.major_breaks_pattern = f"((?:{pattern_str})+(?:\\s+|$))"

        if self.major_breaks_pattern is not None:
            self.major_breaks_pattern = maybe_compile_regex(self.major_breaks_pattern)

        # Minor breaks (don't split sentences)
        if (self.minor_breaks_pattern is None) and self.minor_breaks:
            pattern_str = "|".join(re.escape(b) for b in self.minor_breaks)

            # Match minor break with either whitespace at the end or at the end of the text
            self.minor_breaks_pattern = f"((?:{pattern_str})(?:\\s+|$))"

        if self.minor_breaks_pattern is not None:
            self.minor_breaks_pattern = maybe_compile_regex(self.minor_breaks_pattern)

        # Word breaks (break words apart into multiple words)
        if (self.word_breaks_pattern is None) and self.word_breaks:
            pattern_str = "|".join(re.escape(b) for b in self.word_breaks)
            self.word_breaks_pattern = f"(?:{pattern_str})"

        if self.word_breaks_pattern is not None:
            self.word_breaks_pattern = maybe_compile_regex(self.word_breaks_pattern)

        # Currency
        if not self.currencies:
            try:
                # Look up currencies for locale
                locale_obj = babel.Locale(self.babel_locale)

                # $ -> USD
                self.currencies = {
                    babel.numbers.get_currency_symbol(cn): cn
                    for cn in locale_obj.currency_symbols
                }
            except Exception:
                # No automatic currencies
                pass

        if not self.currency_symbols:
            # Currency symbols (e.g., "$") by decreasing length
            self.currency_symbols = sorted(
                self.currencies, key=operator.length_hint, reverse=True
            )


# -----------------------------------------------------------------------------



#!/usr/bin/env python3
"""Guess word pronunciations using a Phonetisaurus FST

See bin/fst2npz.py to convert an FST to a numpy graph.
"""
import argparse
import logging
import os
import sys
import time
import typing
from collections import defaultdict
from pathlib import Path

import numpy as np

_LOGGER = logging.getLogger("g2p_phonetisaurus")

NUMPY_GRAPH = typing.Dict[str, np.ndarray]

# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="g2p_phonetisaurus")

    # Create subparsers for each sub-command
    sub_parsers = parser.add_subparsers()
    sub_parsers.required = True
    sub_parsers.dest = "command"

    # -------
    # Predict
    # -------
    predict_parser = sub_parsers.add_parser(
        "predict", help="Predict phonemes for word(s)"
    )
    predict_parser.add_argument(
        "--graph", required=True, help="Path to graph npz file from fst2npy.py"
    )
    predict_parser.add_argument(
        "words", nargs="*", help="Words to guess pronunciations for"
    )
    predict_parser.add_argument(
        "--max-guesses",
        default=1,
        type=int,
        help="Maximum number of guesses per word (default: 1)",
    )
    predict_parser.add_argument(
        "--beam",
        default=500,
        type=int,
        help="Initial width of search beam (default: 500)",
    )
    predict_parser.add_argument(
        "--min-beam",
        default=100,
        type=int,
        help="Minimum width of search beam (default: 100)",
    )
    predict_parser.add_argument(
        "--beam-scale",
        default=0.6,
        type=float,
        help="Scalar multiplied by beam after each step (default: 0.6)",
    )
    predict_parser.add_argument(
        "--grapheme-separator",
        default="",
        help="Separator between input graphemes (default: none)",
    )
    predict_parser.add_argument(
        "--phoneme-separator",
        default=" ",
        help="Separator between output phonemes (default: space)",
    )
    predict_parser.add_argument(
        "--preload-graph",
        action="store_true",
        help="Preload graph into memory before starting",
    )
    predict_parser.set_defaults(func=do_predict)

    # ----
    # Test
    # ----
    test_parser = sub_parsers.add_parser("test", help="Test G2P model on a lexicon")
    test_parser.add_argument(
        "--graph", required=True, help="Path to graph npz file from fst2npy.py"
    )
    test_parser.add_argument(
        "texts", nargs="*", help="Lines with '<word> <phoneme> <phoneme> ...'"
    )
    test_parser.add_argument(
        "--beam",
        default=500,
        type=int,
        help="Initial width of search beam (default: 500)",
    )
    test_parser.add_argument(
        "--min-beam",
        default=100,
        type=int,
        help="Minimum width of search beam (default: 100)",
    )
    test_parser.add_argument(
        "--beam-scale",
        default=0.6,
        type=float,
        help="Scalar multiplied by beam after each step (default: 0.6)",
    )
    test_parser.add_argument(
        "--preload-graph",
        action="store_true",
        help="Preload graph into memory before starting",
    )
    test_parser.set_defaults(func=do_test)

    # ----------------
    # Shared arguments
    # ----------------
    for sub_parser in [predict_parser, test_parser]:
        sub_parser.add_argument(
            "--debug", action="store_true", help="Print DEBUG messages to console"
        )

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    args.func(args)


# -----------------------------------------------------------------------------


def do_predict(args):
    """Predict phonemes for words"""
    args.graph = Path(args.graph)

    _LOGGER.debug("Loading graph from %s", args.graph)
    phon_graph = PhonetisaurusGraph.load(args.graph, preload=args.preload_graph)

    if args.words:
        # Arguments
        words = args.words
        _LOGGER.info("Guessing pronunciations for %s word(s)", len(words))
    else:
        # Standard input
        words = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading words from stdin...", file=sys.stderr)

    # Guess pronunciations
    for word, graphemes, phonemes in phon_graph.g2p(
        words,
        grapheme_separator=args.grapheme_separator,
        max_guesses=args.max_guesses,
        beam=args.beam,
        min_beam=args.min_beam,
        beam_scale=args.beam_scale,
    ):
        if not phonemes:
            _LOGGER.warning("No pronunciation for %s (%s)", word, graphemes)
            continue

        print(word, args.phoneme_separator.join(phonemes))


# -----------------------------------------------------------------------------


def do_test(args):
    """Test performance relative a known lexicon"""
    try:
        from rapidfuzz.string_metric import levenshtein
    except ImportError as e:
        _LOGGER.critical("rapidfuzz library is needed for levenshtein distance")
        _LOGGER.critical("pip install 'rapidfuzz>=1.4.1'")
        raise e

    args.graph = Path(args.graph)

    _LOGGER.debug("Loading graph from %s", args.graph)
    phon_graph = PhonetisaurusGraph.load(args.graph, preload=args.preload_graph)

    if args.texts:
        lines = args.texts
    else:
        lines = sys.stdin

        if os.isatty(sys.stdin.fileno()):
            print("Reading lexicon lines from stdin...", file=sys.stderr)

    # Load lexicon
    lexicon = {}
    for line in lines:
        line = line.strip()
        if (not line) or (" " not in line):
            continue

        word, actual_phonemes = line.split(maxsplit=1)
        lexicon[word] = actual_phonemes

    # Predict phonemes
    predicted_phonemes = {}
    start_time = time.perf_counter()

    for word in lexicon:
        for _, _, guessed_phonemes in phon_graph.g2p(
            [word],
            beam=args.beam,
            min_beam=args.min_beam,
            beam_scale=args.beam_scale,
            max_guesses=1,
        ):
            predicted_phonemes[word] = " ".join(guessed_phonemes)

            # Only one guess
            break

    end_time = time.perf_counter()

    # Calculate PER
    num_errors = 0
    num_missing = 0
    num_phonemes = 0

    for word, actual_phonemes in lexicon.items():
        expected_phonemes = predicted_phonemes.get(word, "")

        if expected_phonemes:
            distance = levenshtein(expected_phonemes, actual_phonemes)
            num_errors += distance
            num_phonemes += len(actual_phonemes)
        else:
            num_missing += 1
            _LOGGER.warning("No pronunciation for %s", word)

    assert num_phonemes > 0, "No phonemes were read"

    # Calculate results
    per = round(num_errors / num_phonemes, 2)
    wps = round(len(predicted_phonemes) / (end_time - start_time), 2)
    print("PER:", per, "Errors:", num_errors, "words/sec:", wps)

    if num_missing > 0:
        print("Total missing:", num_missing)


# -----------------------------------------------------------------------------

_NOT_FINAL = object()


class PhonetisaurusGraph:
    """Graph of numpy arrays that represents a Phonetisaurus FST

    Also contains shared cache of edges and final state probabilities.
    These caches are necessary to ensure that the .npz file stays small and fast
    to load.
    """

    def __init__(self, graph: NUMPY_GRAPH, preload: bool = False):
        self.graph = graph

        self.start_node = int(self.graph["start_node"].item())

        # edge_index -> (from_node, to_node, ilabel, olabel)
        self.edges = self.graph["edges"]
        self.edge_probs = self.graph["edge_probs"]

        # int -> [str]
        self.symbols = []
        for symbol_str in self.graph["symbols"]:
            symbol_list = symbol_str.replace("_", "").split("|")
            self.symbols.append((len(symbol_list), symbol_list))

        # nodes that are accepting states
        self.final_nodes = self.graph["final_nodes"]

        # node -> probability
        self.final_probs = self.graph["final_probs"]

        # Cache
        self.preloaded = preload
        self.out_edges: typing.Dict[int, typing.List[int]] = defaultdict(list)
        self.final_node_probs: typing.Dict[int, typing.Any] = {}

        if preload:
            # Load out edges
            for edge_idx, (from_node, *_) in enumerate(self.edges):
                self.out_edges[from_node].append(edge_idx)

            # Load final probabilities
            self.final_node_probs.update(zip(self.final_nodes, self.final_probs))

    @staticmethod
    def load(graph_path: typing.Union[str, Path], **kwargs) -> "PhonetisaurusGraph":
        """Load .npz file with numpy graph"""
        np_graph = np.load(graph_path, allow_pickle=True)
        return PhonetisaurusGraph(np_graph, **kwargs)

    def g2p(
        self, words: typing.Iterable[typing.Union[str, typing.Sequence[str]]], **kwargs
    ) -> typing.Iterable[
        typing.Tuple[
            typing.Union[str, typing.Sequence[str]],
            typing.Sequence[str],
            typing.Sequence[str],
        ],
    ]:
        """Guess phonemes for words"""
        for word in words:
            for graphemes, phonemes in self.g2p_one(word, **kwargs):
                yield word, graphemes, phonemes

    def g2p_one(
        self,
        word: typing.Union[str, typing.Sequence[str]],
        eps: str = "<eps>",
        beam: int = 5000,
        min_beam: int = 100,
        beam_scale: float = 0.6,
        grapheme_separator: str = "",
        max_guesses: int = 1,
    ) -> typing.Iterable[typing.Tuple[typing.Sequence[str], typing.Sequence[str]]]:
        """Guess phonemes for word"""
        current_beam = beam
        graphemes: typing.Sequence[str] = []

        if isinstance(word, str):
            word = word.strip()

            if grapheme_separator:
                graphemes = word.split(grapheme_separator)
            else:
                graphemes = list(word)
        else:
            graphemes = word

        if not graphemes:
            return graphemes, []

        # (prob, node, graphemes, phonemes, final, beam)
        q: typing.List[
            typing.Tuple[
                float,
                typing.Optional[int],
                typing.Sequence[str],
                typing.List[str],
                bool,
            ]
        ] = [(0.0, self.start_node, graphemes, [], False)]

        q_next: typing.List[
            typing.Tuple[
                float,
                typing.Optional[int],
                typing.Sequence[str],
                typing.List[str],
                bool,
            ]
        ] = []

        # (prob, phonemes)
        best_heap: typing.List[typing.Tuple[float, typing.Sequence[str]]] = []

        # Avoid duplicate guesses
        guessed_phonemes: typing.Set[typing.Tuple[str, ...]] = set()

        while q:
            done_with_word = False
            q_next = []

            for prob, node, next_graphemes, output, is_final in q:
                if is_final:
                    # Complete guess
                    phonemes = tuple(output)
                    if phonemes not in guessed_phonemes:
                        best_heap.append((prob, phonemes))
                        guessed_phonemes.add(phonemes)

                    if len(best_heap) >= max_guesses:
                        done_with_word = True
                        break

                    continue

                assert node is not None

                if not next_graphemes:
                    if self.preloaded:
                        final_prob = self.final_node_probs.get(node, _NOT_FINAL)
                    else:
                        final_prob = self.final_node_probs.get(node)
                        if final_prob is None:
                            final_idx = int(np.searchsorted(self.final_nodes, node))
                            if self.final_nodes[final_idx] == node:
                                # Cache
                                final_prob = float(self.final_probs[final_idx])
                                self.final_node_probs[node] = final_prob
                            else:
                                # Not a final state
                                final_prob = _NOT_FINAL
                                self.final_node_probs[node] = final_prob

                    if final_prob != _NOT_FINAL:
                        final_prob = typing.cast(float, final_prob)
                        q_next.append((prob + final_prob, None, [], output, True))

                len_next_graphemes = len(next_graphemes)
                if self.preloaded:
                    # Was pre-loaded in __init__
                    edge_idxs = self.out_edges[node]
                else:
                    # Build cache during search
                    maybe_edge_idxs = self.out_edges.get(node)
                    if maybe_edge_idxs is None:
                        edge_idx = int(np.searchsorted(self.edges[:, 0], node))
                        edge_idxs = []
                        while self.edges[edge_idx][0] == node:
                            edge_idxs.append(edge_idx)
                            edge_idx += 1

                        # Cache
                        self.out_edges[node] = edge_idxs
                    else:
                        edge_idxs = maybe_edge_idxs

                for edge_idx in edge_idxs:
                    _, to_node, ilabel_idx, olabel_idx = self.edges[edge_idx]
                    out_prob = self.edge_probs[edge_idx]

                    len_igraphemes, igraphemes = self.symbols[ilabel_idx]

                    if len_igraphemes > len_next_graphemes:
                        continue

                    if igraphemes == [eps]:
                        item = (prob + out_prob, to_node, next_graphemes, output, False)
                        q_next.append(item)
                    else:
                        sub_graphemes = next_graphemes[:len_igraphemes]
                        if igraphemes == sub_graphemes:
                            _, olabel = self.symbols[olabel_idx]
                            item = (
                                prob + out_prob,
                                to_node,
                                next_graphemes[len(sub_graphemes) :],
                                output + olabel,
                                False,
                            )
                            q_next.append(item)

            if done_with_word:
                break

            q_next = sorted(q_next, key=lambda item: item[0])[:current_beam]
            q = q_next

            current_beam = max(min_beam, (int(current_beam * beam_scale)))

        # Yield guesses
        if best_heap:
            for _, guess_phonemes in sorted(best_heap, key=lambda item: item[0])[
                :max_guesses
            ]:
                yield graphemes, [p for p in guess_phonemes if p]
        else:
            # No guesses
            yield graphemes, []


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()
